# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging

import time

import datetime
import docker
import pytz
from docker.types import SecretReference, RestartPolicy

from benchsuite.scheduler.config import BenchsuiteSchedulerConfig
from benchsuite.scheduler.schedules import BenchmarkingScheduleConfig

logger = logging.getLogger(__name__)


class BenchsuiteInstance(object):

    def __init__(self, docker_service):
        self.id = docker_service.name
        self.docker_service_id = docker_service.id
        if 'ContainerID' in docker_service.tasks()[0]['Status']['ContainerStatus']:
            self.docker_container_id = docker_service.tasks()[0]['Status']['ContainerStatus']['ContainerID']
        self.status = docker_service.tasks()[0]['Status']['State']

        try:
            self.created = datetime.datetime.strptime(docker_service.tasks()[0]['CreatedAt'][:19], '%Y-%m-%dT%H:%M:%S')
            self.created = self.created.replace(tzinfo=pytz.utc)  # docker api returns dates in UTC
        except:
            logger.error('error parsing data: ', docker_service.tasks()[0]['CreatedAt'][:19])
            pass

        self.schedule_id = docker_service.attrs['Spec']['Labels']['benchsuite.schedule_id']
        self.username = docker_service.attrs['Spec']['Labels']['benchsuite.username']



class DockerManager(object):
    """
    Wraps the a Docker Swarm and provides methods to manage the Benchsuite
    containers
    """

    def __init__(self, config: BenchsuiteSchedulerConfig):

        self.__client = docker.DockerClient(base_url=config.docker_host)
        self._storage_secret_ref = self.__get_secret_ref(config.docker_storage_secret)
        self._global_tags = config.benchsuite_global_tags or []
        self._global_env = config.docker_global_env or {}
        self._additional_opts = config.benchsuite_additional_opts or []
        self._docker_additional_opts = config.docker_additional_opts or {}
        self._benchsuite_multiexec_image = config.docker_benchsuite_image

    def create_benchsuite_multiexec_instance(self, schedule: BenchmarkingScheduleConfig):
        provider_secret_ref = self.__get_secret_ref(schedule.provider_config_secret)
        restartCond = RestartPolicy(condition='none')

        args = [
            '--storage-config', '/run/secrets/' + self._storage_secret_ref['SecretName'],
            '--provider', '/run/secrets/' + provider_secret_ref['SecretName'],
            '--failonerror',
            '--user', schedule.username
        ]

        for t in self._global_tags:
            args.extend(['--tag', t])

        for t in schedule.tags:
            args.extend(['--tag', t])

        args.extend(self._additional_opts)
        args.extend(schedule.benchsuite_additional_opts or [])
        args.extend(schedule.tests)

        final_env = dict(self._global_env)
        final_env.update(schedule.env)
        env_list = ['{0}={1}'.format(k,v) for k,v in final_env.items()]

        name = schedule.id + '_' + schedule.username
        labels = {
            'benchsuite.source': 'benchsuite-scheduler',
            'benchsuite.schedule_id': schedule.id,
            'benchsuite.username': schedule.username}


        docker_additional_opts = self._docker_additional_opts
        docker_additional_opts.update(schedule.docker_additional_opts)
        logger.debug('Appending additional docker opts:', docker_additional_opts)

        service = self.__client.services.create(
            self._benchsuite_multiexec_image,
            secrets=[self._storage_secret_ref, provider_secret_ref],
            args=args,
            env=env_list,
            restart_policy = restartCond,
            labels=labels,
            name=name,
            **docker_additional_opts
        )

        self.__wait_instance_ready(service)

        return BenchsuiteInstance(service)


    def __wait_instance_ready(self, service):
        retry = 10
        cont_id = None
        while retry and not cont_id:
            service.reload()
            if len(service.tasks()) > 0:
                task = service.tasks()[0]
                if 'ContainerID' in task['Status']['ContainerStatus']:
                    cont_id = task['Status']['ContainerStatus']['ContainerID']
                    break

            retry -= 1
            logger.debug('Container ID not ready yet. Retrying in 10 seconds')
            time.sleep(10)

        if not cont_id:
            raise Exception('Failed to retrieve the container id after 100 seconds')


    def wait(self, instance):
        cont = self.__client.containers.get(instance.docker_container_id)
        retval = cont.wait()
        log = cont.logs().decode()
        return retval, log

    def remove_instance(self, instance):
        self.__client.services.get(instance.docker_service_id).remove()

    def list(self):
        services = self.__client.services.list(filters={'label': 'benchsuite.source=benchsuite-scheduler'})

        return [BenchsuiteInstance(s) for s in services]

    def stop_service(self):
        raise NotImplementedError()

    def stop_all(self):
        raise NotImplementedError()

    def __get_secret_ref(self,name_or_id):
        for s in self.__client.secrets.list():
            if s.id == name_or_id or s.name == name_or_id:
                return SecretReference(s.id, s.name)

        return None



# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging
import threading

from apscheduler.events import EVENT_SCHEDULER_SHUTDOWN, EVENT_ALL
from apscheduler.jobstores.memory import MemoryJobStore
from apscheduler.jobstores.mongodb import MongoDBJobStore
from apscheduler.schedulers.background import BackgroundScheduler

from benchsuite.scheduler.dockermanager import DockerManager
from benchsuite.scheduler.jobs.meta import print_scheduled_jobs_info, \
    sync_scheduled_jobs
from benchsuite.scheduler.logger import JobExecutionLogger
from benchsuite.scheduler.schedules import BenchmarkingSchedulesDB

logger = logging.getLogger(__name__)



# we want to have a global instance of BSScheduler so that jobs can access it
# we do not want to specify the BSScheduler instance as jobs parameters because
# in this case it would be serialized in the jobs status while we want a fresh
# instance if we restart the scheduler
__instance = None


def get_bsscheduler():
    """
    Returns the global instance of the BSScheduler
    :return:
    """
    return __instance


def create_bsscheduler(config):
    """
    Creates a new global instance of the BSSCheduler
    :return:
    """
    global __instance
    __instance = BSScheduler(config)
    return get_bsscheduler()






class BSScheduler(object):
    """
    This class wraps the information that might be needed to the Jobs
    """

    config = None
    scheduler = None
    schedulesdb = None
    jobslogger = None
    dockermanager = None

    __initialized = False

    def __init__(self, config):
        self.config = config
        self.__blocked_flag = None

    def initialize(self):
        """
        Initialize the scheduler
        :return:
        """

        # load the schedules configuration
        self.schedules_db = BenchmarkingSchedulesDB(
            self.config.schedules_db_host,
            self.config.schedules_db_name,
            self.config.schedules_db_collection)

        self.dockermanager = DockerManager(self.config)

        # initialize the APScheduler
        jobstore = MongoDBJobStore(
            database=self.config.jobs_db_name,
            collection=self.config.jobs_db_collection,
            host=self.config.jobs_db_host
        )

        self.scheduler = BackgroundScheduler(jobstores={
            'benchmarking_jobs':jobstore,
            'meta_jobs': MemoryJobStore()})

        logger.debug('New APScheduler created: ' + str(self.scheduler))

        # if we do not start the scheduler, the jobstores are not initialized
        # and the old jobs (from the DB) are not visible (they are needed
        # because we want to sync the store before starting to execute the jobs)
        self.scheduler.start(paused=True)

        self.__add_meta_jobs()

        # initialize the job logger facility
        self.jobslogger = JobExecutionLogger(
            self.config.exec_db_host,
            self.config.exec_db_name,
            self.config.exec_db_collection,
            log_missed_execution=False
        )
        self.scheduler.add_listener(
            self.jobslogger.apscheduler_listener, EVENT_ALL)

        # run the sync at the beginning
        sync_scheduled_jobs(self)

        self.__initialized = True

        logger.info('Printing jobs info. It will printed every minute and '
                    'everytime the job list is modified')
        print_scheduled_jobs_info(self)


    def __add_meta_jobs(self):
        j = self.scheduler.add_job(sync_scheduled_jobs, 'interval',
                               seconds=self.config.schedules_sync_interval,
                               id='_sync_schedules', replace_existing=True,
                               max_instances=1,
                               args=[self],
                               jobstore='meta_jobs')
        logger.debug('Added meta job to sync schedules. Next run: %s', str(j.next_run_time))

        j = self.scheduler.add_job(print_scheduled_jobs_info,
                               'interval', seconds=self.config.print_jobs_info_interval,
                               id='_print_jobs_info', replace_existing=True,
                               max_instances=1,
                               args=[self],
                               jobstore='meta_jobs')
        logger.debug('Added meta job to print jobs info. Next run: %s', str(j.next_run_time))

    def start(self):
        """
        Starts to process jobs. This method also initializes the scheduler is
        not done
        :return:
        """
        if not self.__initialized:
            logger.warning('The scheduler is not initialized. Initializing it'
                           'before starting')
            self.initialize()

        logger.info('String processing the jobs')
        self.scheduler.resume()

    def shutdown(self):
        """
        Stops the scheduler
        :return:
        """
        logger.info('Shutting down the scheduler')
        self.scheduler.shutdown()

    def wait_unitl_shutdown(self):
        """
        Blocking call to wait until the scheduler is shutoff
        :return:
        """
        self.scheduler.add_listener(self.__scheduler_listener,
                                    EVENT_SCHEDULER_SHUTDOWN)

        self.__blocked_flag = threading.Condition()

        with self.__blocked_flag:
            self.__blocked_flag.wait()

    def __scheduler_listener(self, event):
        if event.code == EVENT_SCHEDULER_SHUTDOWN:
            with self.__blocked_flag:
                self.__blocked_flag.notify()











# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging

from pymongo import MongoClient

logger = logging.getLogger(__name__)

_DEFAULT_INTERVAL = {
    'weeks': 0,
    'days': 0,
    'hours': 0,
    'minutes': 0,
    'seconds': 0
    }

class BenchmarkingScheduleConfig(object):

    interval = None
    username = None
    tags = []
    tests = None
    env = {}
    provider_config_secret = None
    benchsuite_additional_opts = []
    docker_additional_opts = {}

    def __init__(self, raw_obj):
        self._raw_obj = raw_obj

        self.id = raw_obj['id']

        raw_interval = raw_obj['interval']
        try:
            self.interval = {
                k:int(raw_interval[k]) if k in raw_interval else v
                for k, v in _DEFAULT_INTERVAL.items()}
        except ValueError:
            logger.error('Error parsing schedule interval. All interval '
                         'values must be integer')
            raise

        self.provider_config_secret = raw_obj['provider_config_secret']
        self.username = raw_obj['username']

        self.tests = raw_obj['tests']

        # optional
        if 'tags' in raw_obj:
            self.tags = raw_obj['tags']

        if 'env' in raw_obj:
            self.env = raw_obj['env']

        if 'benchsuite_additional_opts' in raw_obj:
            self.benchsuite_additional_opts = raw_obj['benchsuite_additional_opts']

        if 'docker_additional_opts' in raw_obj:
            self.docker_additional_opts = raw_obj['docker_additional_opts']

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

class BenchmarkingSchedulesDB(object):


    def __init__(self, db_host, db_name, collection):
        self._client = MongoClient(db_host)
        self._db_name = db_name
        self._collection = collection
        logger.info('Schedules DB at {0}, db_name={1}, collection={2}'.format(
            db_host, db_name, collection))

    def get_benchmarking_schedules(self):
        out = []
        for r in self._client[self._db_name][self._collection].find():

            if 'active' not in r or not r['active']:
                continue

            try:
                out.append(BenchmarkingScheduleConfig(r))
            except:
                logger.error('Schedule %s malformed. Not considering it',
                             r['id'])

        return out



# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging

from apscheduler.events import EVENT_JOB_EXECUTED, EVENT_JOB_ERROR, \
    EVENT_JOB_MISSED, EVENT_JOB_ADDED, EVENT_JOB_REMOVED, EVENT_JOB_MODIFIED
from pymongo import MongoClient

from benchsuite.scheduler.jobs.meta import print_scheduled_jobs_info
from benchsuite.scheduler.synchronizer import JOB_ID_PREFIX

logger = logging.getLogger(__name__)

class JobExecutionLogger(object):
    """
    provides methods to log the execution of jobs in the console and in a
    MongoDB collection
    """

    _client = None

    _log_success_execution = None
    _log_error_execution = None
    _log_missed_execution = None

    def __init__(self, db_host, db_name, db_collection,
                 log_success_execution=False,
                 log_error_execution=True,
                 log_missed_execution=True):

        self._client = MongoClient(db_host)
        self._db_name = db_name
        self._db_collection = db_collection
        self._log_success_execution = log_success_execution
        self._log_error_execution = log_error_execution
        self._log_missed_execution = log_missed_execution

    def apscheduler_listener(self, evt):

        if evt.code == EVENT_JOB_EXECUTED:
            logger.info('[LOG] Job successfully executed')
            if self._log_success_execution:
                self.__add_entry('OK', evt)

        elif evt.code == EVENT_JOB_ERROR:
            logger.info('[LOG] Job ERROR')
            if self._log_error_execution:
                self.__add_entry('ERROR', evt)

        elif evt.code == EVENT_JOB_MISSED:
            logger.info('[LOG] Job MISSED')
            if self._log_missed_execution:
                self.__add_entry('MISSSED', evt)


        elif evt.code == EVENT_JOB_ADDED or \
            evt.code == EVENT_JOB_REMOVED or \
            evt.code == EVENT_JOB_MODIFIED:
                logger.info('Job list modified. Printing list')
                from benchsuite.scheduler.bsscheduler import get_bsscheduler
                print_scheduled_jobs_info(get_bsscheduler())

        else:
            logger.debug('Received event code %s that is not managed', evt.code)

    def __add_entry(self, status, evt):
        obj = {
            'status': status,
            'time': evt.scheduled_run_time,
            'ap_job_id': evt.job_id,
        }

        if evt.job_id.startswith(JOB_ID_PREFIX):
            obj['schedule_id'] = evt.job_id[len(JOB_ID_PREFIX):]

        if status == 'OK':
            obj['retval'] = evt.retval

        if status == 'ERROR':
            obj['exception'] = str(evt.exception)
            if hasattr(evt.exception, 'log'):
                obj['error_log'] = evt.exception.log

            obj['traceback'] = evt.traceback

        self._client[self._db_name][self._db_collection].insert_one(obj)






# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging
from datetime import timedelta, datetime
from typing import List

from apscheduler.job import Job
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.triggers.interval import IntervalTrigger

from benchsuite.scheduler.schedules import BenchmarkingScheduleConfig, \
    BenchmarkingSchedulesDB

logger = logging.getLogger(__name__)


JOB_ID_PREFIX = 'ap-'

def __build_args_list(schedule):
    args = [schedule.username,
            schedule.provider_config_secret,
            schedule.tests]

    kwargs = {}

    if schedule.tags:
        kwargs['tags'] = schedule.tags

    if schedule.env:
        kwargs['env'] = schedule.env

    if schedule.additional_opts:
        kwargs['additional_opts'] = schedule.additional_opts

    return args, kwargs

def schedule_new_job(
        schedule: BenchmarkingScheduleConfig,
        scheduler: BaseScheduler):
    """
    schedule a new jobs starting from a schedule
    :param schedule:
    :param scheduler:
    :return:
    """

    assert schedule.interval, '"interval" field in the schedule cannot be None'

    id = JOB_ID_PREFIX + schedule.id
    trigger = IntervalTrigger(**schedule.interval)

    # import here to avoid import loops
    from benchsuite.scheduler.jobs.benchmarking import benchmarking_job

    j = scheduler.add_job(benchmarking_job, trigger=trigger, id=id,
                          args=[schedule],
                          jobstore='benchmarking_jobs',
                          max_instances=1,
                          coalesce=True,
                          # run immediately
                          next_run_time=datetime.now())

    logger.info('[ADD] Job %s with id %s added '
                'for schedule %s', str(j), j.id, schedule.id)


def update_job(job: Job, schedule: BenchmarkingSchedulesDB):
    """
    updates the scheduling interval and parameters of a scheduled jobs based on
    updated data in a schedule. The update is performed only if needed
    :param job:
    :param schedule:
    :return:
    """

    # update the interval if it is changed
    if job.trigger.interval != timedelta(**schedule.interval):
        logger.debug('[UPDATE] Job %s updated because interval '
                     'changed from "%s" to "%s"', job.id, job.trigger.interval,
                     timedelta(**schedule.interval))

        new_trigger = IntervalTrigger(**schedule.interval)
        job.reschedule(new_trigger)


    # update the parameters if changed
    if job.args[0] != schedule:
        logger.debug('[UPDATE] Job %s updated because the args '
                     'are changed', job.id)
        job.modify(args=[schedule])


def sync_jobs(
        schedules: List[BenchmarkingScheduleConfig],
        scheduler: BaseScheduler):
    """
    Syncrhonizes the scheduled jobs with the benchmarking schedules defined in
    the schedules database
    :param schedules:
    :param scheduler:
    :return:
    """

    logger.debug('Synchronizing benchmarking schedules with scheduler jobs')

    # convert from list to dict for easier access
    schedules = {s.id: s for s in schedules}

    # get all the jobs. Filter out jobs with an id not recognized as created
    # from the schedule
    jobs = {j.id[len(JOB_ID_PREFIX):]: j
            for j in scheduler.get_jobs() if j.id.startswith(JOB_ID_PREFIX)}

    logger.debug('Considering %s schedules and %s jobs',
                 len(schedules), len(jobs))

    # for all scheduled jobs...
    for id in jobs.keys():

        if id in schedules:
            # the scheduled jobs appears in the schedules db.
            # check if the parameters are the same otherwise, update it
            # (do not update always to avoid rescheduling and calls to the db)
            update_job(jobs[id], schedules[id])

        else:
            # the scheduled jobs does not exist in the schedules db
            # this means that it has been deleted from the schedules db and must
            # be deleted from here also
            logger.info('[REMOVE] Job %s removed '
                        'becasue not found in the schedules db', id)
            jobs[id].remove()

    # for all schedules...
    for id in schedules.keys():

        if id not in jobs:
            # the schedule does not appear in the scheduler jobs, it must be
            # added
            schedule_new_job(schedules[id], scheduler)

    logger.debug('Syncrhonization completed')


# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging

import datetime
import docker
import pytz
from tzlocal import get_localzone

from benchsuite.scheduler.synchronizer import sync_jobs

logger = logging.getLogger(__name__)

def sync_scheduled_jobs(bsscheduler):
    """
    This function is scheduled as jobs and periodically
    syncrhonize the jobs in the scheduler with the schedules db
    """

    schedules = bsscheduler.schedules_db.get_benchmarking_schedules()

    sync_jobs(schedules, bsscheduler.scheduler)

    return 0


def print_scheduled_jobs_info(bsscheduler):
    """
    print some information on the status of the scheduler
    :param bsscheduler:
    :return:
    """
    out = []
    jobs = bsscheduler.scheduler.get_jobs(jobstore='benchmarking_jobs')

    out.append('********************************************************')
    out.append('*                    SCHEDULED JOBS                    *')
    out.append('*                                                      *')
    for j in jobs:
        delta = j.next_run_time - datetime.datetime.now(tz=j.next_run_time.tzinfo)
        out.append(u'* - {0} \u27A1 {1} (in {2})'.format(j.args[0].id, str(j.next_run_time.strftime('%Y-%m-%d %H:%M:%S %Z')), delta))
    out.append('*                                                      *')
    out.append('*                     RUNNING JOBS                     *')

    all_instances = bsscheduler.dockermanager.list()

    for i in [i for i in all_instances if i.status == 'running']:
        delta = datetime.datetime.now(tz=i.created.tzinfo) - i.created
        out.append('* - {0} {1} (since {2})'.format(i.schedule_id, i.created.strftime('%Y-%m-%d %H:%M:%S %Z'), delta))
    out.append('*                                                      *')
    out.append('*                  NOT RUNNING INSTANCE                *')
    for i in [i for i in all_instances if i.status != 'running']:
        out.append('* - {0} {1}'.format(i.schedule_id, i.created.strftime('%Y-%m-%d %H:%M:%S %Z')))
    out.append('*                                                      *')
    out.append('********************************************************')

    print('\n'.join(out))



# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import json

DEFAULTS = {
    'SCHEDULES_SYNC_INTERVAL': 60,
    'SCHEDULES_JOBS_PRINT_INTERVAL': 60,

    'DB_HOST': None,
    'DB_NAME': 'benchmarking',

    'SCHEDULES_DB_HOST': None,
    'SCHEDULES_DB_NAME': None,
    'SCHEDULES_DB_COLLECTION': 'scheduling',

    'JOBS_DB_HOST': None,
    'JOBS_DB_NAME': None,
    'JOBS_DB_COLLECTION': '_apjobs',

    'EXEC_DB_HOST': None,
    'EXEC_DB_NAME': None,
    'EXEC_DB_COLLECTION': '_apexec',

    'DOCKER_HOST': 'localhost:2375',
    'DOCKER_STORAGE_SECRET': None,
    'DOCKER_BENCHSUITE_IMAGE': 'benchsuite/benchsuite-multiexec',
    'DOCKER_GLOBAL_ENV': '',
    'DOCKER_ADDITIONAL_OPTS': '',
    'BENCHSUITE_GLOBAL_TAGS': '',
    'BENCHSUITE_ADDITIONAL_OPTS': ''
}

class BenchsuiteSchedulerConfig(object):
    """the global configuration for the Benchsuite Scheduler"""

    def __init__(self, user_config):

        if 'DB_HOST' not in user_config and (
                        'SCHEDULES_DB_HOST' not in user_config or
                        'JOBS_DB_HOST' not in user_config or
                        'EXEC_DB_HOST' not in user_config):
            raise ValueError('Either DB_HOST or SCHEDULES_DB_HOST, '
                             'JOBS_DB_HOST and EXEC_DB_HOST are mandatory')

        if 'DOCKER_STORAGE_SECRET' not in user_config:
            raise ValueError('The DOCKER_STORAGE_SECRET config parameter '
                             'is mandatory.')

        # load defaults
        cfg = dict(DEFAULTS)

        # override with user values
        cfg.update(user_config)

        self.schedules_sync_interval = cfg['SCHEDULES_SYNC_INTERVAL']
        self.print_jobs_info_interval = cfg['SCHEDULES_JOBS_PRINT_INTERVAL']

        self.schedules_db_host = cfg['SCHEDULES_DB_HOST'] or cfg['DB_HOST']
        self.schedules_db_name = cfg['SCHEDULES_DB_NAME'] or cfg['DB_NAME']
        self.schedules_db_collection = cfg['SCHEDULES_DB_COLLECTION']

        self.jobs_db_host = cfg['JOBS_DB_HOST'] or cfg['DB_HOST']
        self.jobs_db_name = cfg['JOBS_DB_NAME'] or cfg['DB_NAME']
        self.jobs_db_collection = cfg['JOBS_DB_COLLECTION']

        self.exec_db_host = cfg['EXEC_DB_HOST'] or cfg['DB_HOST']
        self.exec_db_name = cfg['EXEC_DB_NAME'] or cfg['DB_NAME']
        self.exec_db_collection = cfg['EXEC_DB_COLLECTION']

        self.docker_host = cfg['DOCKER_HOST']
        self.docker_storage_secret = cfg['DOCKER_STORAGE_SECRET']
        self.docker_benchsuite_image = cfg['DOCKER_BENCHSUITE_IMAGE']

        self.docker_global_env = {}
        if cfg['DOCKER_GLOBAL_ENV']:

            # the '\,' sequence must be interpreted as a real comma and not as
            # the separator between variables. So first we temporarly replace it
            # with a different sequence (i.e. '\@') and then, after the
            # splitting, change it back to ','
            temp = cfg['DOCKER_GLOBAL_ENV'].replace('\,', '\@')
            for i in temp.split(','):
                t = i.split('=')
                self.docker_global_env[t[0]] = t[1].replace('\@',',')

        self.benchsuite_global_tags = []
        if cfg['BENCHSUITE_GLOBAL_TAGS']:
            self.benchsuite_global_tags = [i for i in cfg['BENCHSUITE_GLOBAL_TAGS'].split(',')]

        self.benchsuite_additional_opts = cfg['BENCHSUITE_ADDITIONAL_OPTS'].split() or []

        self.docker_additional_opts = {}
        if cfg['DOCKER_ADDITIONAL_OPTS']:
            for opt in cfg['DOCKER_ADDITIONAL_OPTS'].split(','):
                kstr, vstr = opt.split("=", maxsplit=1)
                val = json.loads(vstr)
                self.docker_additional_opts[kstr] = val


# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import argparse
import logging
import signal
import time
import sys

import os

from benchsuite.scheduler import config
from benchsuite.scheduler.bsscheduler import get_bsscheduler, create_bsscheduler
from benchsuite.scheduler.config import BenchsuiteSchedulerConfig

logger = logging.getLogger(__name__)

def on_exit(sig, func=None):
    get_bsscheduler().shutdown()
    print('Bye bye...')
    sys.exit(1)



def get_config_parameters(config_file, use_env=True):

    cfg = {}
    if config_file:

        # config file format:
        # PARAM = VALUE

        with open(config_file) as f:
            lines = f.readlines()

        for l in lines:
            tokens = l.split('=')
            cfg[tokens[0].strip()] = tokens[1].strip()

    if use_env:

        for k in config.DEFAULTS.keys():
            if k in os.environ:
                cfg[k] = os.environ[k]

    return cfg




def main(args=None):

    signal.signal(signal.SIGINT, on_exit)

    logging.basicConfig(
        format='%(asctime)s - %(levelname)-7s %(message)s',
        level=logging.DEBUG,
        stream=sys.stdout)

    logging.getLogger('apscheduler').setLevel(logging.WARN)

    logger.info('Logging configured')


    parser = argparse.ArgumentParser(prog='benchsuite-scheduler')
    parser.add_argument('--config', '-c', type=str, help='the location of the config file')


    args = parser.parse_args(args = args)

    if args.config and not os.path.isfile(args.config):
        logger.info('Configuration file {0} does not exist. Do not considering it'.format(args.config))
        args.config = None

    config = get_config_parameters(args.config)



    bsscheduler = create_bsscheduler(BenchsuiteSchedulerConfig(config))
    bsscheduler.initialize()
    bsscheduler.start()

    # wait for the APScheduler thread finishes. This will happen when the
    # scheduler is shut down from a SIGINT (see on_exit function)
    #bsscheduler.scheduler._thread.join()

    bsscheduler.wait_unitl_shutdown()


if __name__ == '__main__':
    main(sys.argv[1:])


# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)
import logging
from typing import Any

from benchsuite.scheduler.bsscheduler import get_bsscheduler
from benchsuite.scheduler.schedules import BenchmarkingScheduleConfig

logger = logging.getLogger(__name__)




class DockerJobFailedException(Exception):

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        self.retval = None
        self.log = None


def benchmarking_job(schedule: BenchmarkingScheduleConfig):

    logger.debug('Starting a benchmarking job')

    dockermanager = get_bsscheduler().dockermanager

    instance = dockermanager.create_benchsuite_multiexec_instance(schedule)

    retval, log = dockermanager.wait(instance)

    dockermanager.remove_instance(instance)

    if retval != 0:
        e = DockerJobFailedException(
            'The execution exit with status {0}'.format(retval))
        e.log = log
        raise e

    return retval




# Benchmarking Suite
# Copyright 2014-2017 Engineering Ingegneria Informatica S.p.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Developed in the ARTIST EU project (www.artist-project.eu) and in the
# CloudPerfect EU project (https://cloudperfect.eu/)

import os
from setuptools import setup, find_packages

# import the VERSION from the source code
import sys
sys.path.append(os.getcwd() + '/src/benchsuite')
from scheduler import VERSION

setup(
    name='benchsuite.scheduler',
    version='.'.join(map(str, VERSION)),

    description='A scheduler to start Benchmarking Suite jobs at defined intervals',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),

    url='https://github.com/benchmarking-suite/benchsuite-scheduler',

    author='Gabriele Giammatteo',
    author_email='gabriele.giammatteo@eng.it',

    license='Apache',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: System :: Benchmark',
        'Topic :: Utilities',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Environment :: Console',
        'Operating System :: Unix',
        'Framework :: Flask'
    ],
    keywords='benchmarking cloud testing performance',

    packages=find_packages('src'),
    namespace_packages=['benchsuite'],
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': ['benchsuite-scheduler=benchsuite.scheduler.cli:main'],
    },

    install_requires=['apscheduler', 'pymongo', 'docker']

)
